package duso.business;

import java.util.*;

import duso.data.DataMatrix;
import duso.data.DataPoint;
import duso.data.IndexMatrix;
import duso.data.Subspace;

public class DUSOProcessing 
{
	public static double computeNormalizedCMOIMI(DataMatrix dataMatrix, IndexMatrix indexMatrix, ArrayList<Integer> dims, double[] CRES, double alpha, Random ranIndexBlock, Random ranSubIndexBlock, int normConstraint, int normType, double upperVal, double[] means, double[] devs) throws Exception
	{
		double cmoimi;
		Subspace tmp = computeCEWithOrder(dataMatrix, indexMatrix, dims, CRES, alpha, ranIndexBlock, means, devs);
		cmoimi = tmp.contrast;
		return cmoimi;
	}
	
	@SuppressWarnings("unchecked")
	public static double computeODMI(DataMatrix dataMatrix, IndexMatrix indexMatrix, ArrayList<Integer> dims, double CRES, double alpha, Random ranIndexBlock, Random ranSubIndexBlock) throws Exception
	{
		double odmi = CRES;
		int num_dims = dims.size();
		
		ArrayList<Integer>[] IN;
		
		int[] subBlockSizes = new int[num_dims - 1];
		int[] maxSubBlockPos = new int[num_dims - 1];
		int maxSize = (int)Math.ceil(dataMatrix.rows * Math.pow(alpha, 1.0 / (num_dims - 1)));
		int maxPos = dataMatrix.rows - maxSize;
		for (int i = num_dims - 2; i >= 0; i--)
		{
			subBlockSizes[i] = (int)Math.ceil(dataMatrix.rows * Math.pow(alpha, 1.0 / (i + 1)));
			maxSubBlockPos[i] = maxSize - subBlockSizes[i];
		}
		
		int[] indexBlock = null;
		int currentCols;
		int startPosition;
		int[] subIndexBlock = null;
		int subStartPosition;
		ArrayList<Integer> tmp = null;
		ArrayList<Integer> keepPoints = null;
		boolean keep;
		int size;
		int v;
		
		double sum = 0;
		for (int h = 0; h < Constants.NUM_SUBSAMPLING; h++)
		{
			IN = new ArrayList[num_dims - 1];
			for (int j = 0; j < num_dims - 1; j++)
			{
				currentCols = dims.get(j).intValue();
				
				startPosition = DataProcess.generateRandomNumber(ranIndexBlock, 0, maxPos, false);
				indexBlock = DataProcess.getIndexBlock(indexMatrix, currentCols, startPosition, maxSize);
				
				subIndexBlock = null;
				subStartPosition = -1;
				for (int k = j; k < num_dims - 1; k++)
				{
					subStartPosition = DataProcess.generateRandomNumber(ranSubIndexBlock, 0, maxSubBlockPos[j], false);
					subIndexBlock = DataProcess.getIndexBlock(indexBlock, subStartPosition, subBlockSizes[j]);
					
					if (IN[j] == null)
					{
						IN[j] = new ArrayList<Integer>();
						for (int z = 0; z < subIndexBlock.length; z++)
							IN[j].add(new Integer(subIndexBlock[z]));
					}
					else
					{
						tmp = IN[j];
						keepPoints = new ArrayList<Integer>();
						keep = false;
						size = tmp.size();
						for (int z = 0; z < size; z++)
						{
							keep = false;
							v = tmp.get(z).intValue();
							for (int y = 0; y < subIndexBlock.length; y++)
								if (v == subIndexBlock[y])
								{
									keep = true;
									break;
								}
							
							if (keep == true)
								keepPoints.add(new Integer(v));
						}
						
						IN[j] = keepPoints;
					} // end if
				} // end for
			} // end for
			
			double[] vals = null;
			for (int i = 1; i < num_dims; i++)
			{
				ArrayList<Integer> curIndexBlock = IN[i - 1];
				vals = new double[curIndexBlock.size()];
				for (int j = 0; j < vals.length; j++)
					vals[j] = dataMatrix.data.get(curIndexBlock.get(j).intValue()).measures[dims.get(i).intValue()];
				sum += computeCRE(vals, false, Constants.USE_RAO_CRE);
			}
		}
		
		odmi = odmi - sum / Constants.NUM_SUBSAMPLING;
		return Math.max(odmi, 0.0);
	}
	
	public static Subspace computeCEWithOrder(DataMatrix dataMatrix, IndexMatrix indexMatrix, ArrayList<Integer> dims, double[] CRES, double alpha, Random ranIndexBlock, double[] means, double[] devs) throws Exception
	{
		int numDims = dims.size();
		Subspace ret = new Subspace(numDims);
		
		ArrayList<Integer> tmpDims = new ArrayList<Integer>();
		ArrayList<Integer> tmpFrontDims = new ArrayList<Integer>();
		double maxPairContrast = -Double.MAX_VALUE;
		ArrayList<Integer> maxPair = new ArrayList<Integer>();
		double tmpContrast;
		for (int i = 0; i < numDims; i++)
			for (int j = 0; j < numDims; j++)
				if (j != i)
				{
					tmpDims.clear();
					tmpFrontDims.clear();
					tmpDims.add(new Integer(dims.get(i).intValue()));
					tmpFrontDims.add(new Integer(dims.get(j).intValue()));
					if (Constants.USE_DUSO_SEED == false)
						tmpContrast = computeContrastSpecial(dataMatrix, indexMatrix, tmpDims, tmpFrontDims, CRES, alpha, ranIndexBlock);
					else
						tmpContrast = computeContrastSpecial(dataMatrix, tmpDims, tmpFrontDims, CRES, Constants.NUM_SEEDS, Constants.RAN_SEED_GEN, means, devs);
					
					if (tmpContrast > maxPairContrast)
					{
						maxPairContrast = tmpContrast;
						maxPair.clear();
						maxPair.add(new Integer(tmpDims.get(0).intValue()));
						maxPair.add(new Integer(tmpDims.get(1).intValue()));
					}
				}
		
		int firstDim = maxPair.get(0).intValue();
		int secondDim = maxPair.get(1).intValue();
		ArrayList<Integer> remainingDims = new ArrayList<Integer>();
		int curDim;
		for (int i = 0; i < numDims; i++)
		{
			curDim = dims.get(i).intValue();
			if (curDim != firstDim && curDim != secondDim)
				remainingDims.add(new Integer(i));
		}
		
		int curLength = 2;
		double contrast = maxPairContrast;
		while (curLength < numDims)
		{
			if (Constants.USE_DUSO_SEED == false)
				tmpContrast = computeContrastSpecial(dataMatrix, indexMatrix, maxPair, remainingDims, CRES, alpha, ranIndexBlock);
			else
				tmpContrast = computeContrastSpecial(dataMatrix, maxPair, remainingDims, CRES, Constants.NUM_SEEDS, Constants.RAN_SEED_GEN, means, devs);
			
			contrast += tmpContrast;
			curLength++;
		}
		
		ret.contrast = contrast;
		for (int i = 0; i < numDims; i++)
			ret.dims.add(new Integer(maxPair.get(i).intValue()));
		
		return ret;
	}
	
	public static double computeContrastSpecial(DataMatrix dataMatrix, IndexMatrix indexMatrix, ArrayList<Integer> dims, ArrayList<Integer> frontDims, double[] CRES, double alpha, Random ranIndexBlock) throws Exception
	{
		int numDims = dims.size();
		
		ArrayList<Integer> IN = null;
		int[] indexBlock = null;
		int currentCols;
		int startPosition;
		boolean keep;
		int size;
		ArrayList<Integer> keepPoints = null;
		int v;
		int blockSize = (int)Math.ceil(dataMatrix.rows * Math.pow(alpha, 1.0 / numDims));
		int maxBlockPosition = dataMatrix.rows - blockSize;
		for (int i = 0; i < numDims; i++)
		{
			currentCols = dims.get(i).intValue();
			
			startPosition = DataProcess.generateRandomNumber(ranIndexBlock, 0, maxBlockPosition, false);
			indexBlock = DataProcess.getIndexBlock(indexMatrix, currentCols, startPosition, blockSize);
			
			if (IN == null)
			{
				IN = new ArrayList<Integer>();
				for (int j = 0; j < indexBlock.length; j++)
					IN.add(new Integer(indexBlock[j]));
			}
			else
			{
				keepPoints = new ArrayList<Integer>();
				keep = false;
				size = IN.size();
				for (int j = 0; j < size; j++)
				{
					keep = false;
					v = IN.get(j).intValue();
					for (int y = 0; y < indexBlock.length; y++)
						if (v == indexBlock[y])
						{
							keep = true;
							break;
						}
					
					if (keep == true)
						keepPoints.add(new Integer(v));
				}
				
				IN = keepPoints;
			} // end if
		}
		
		double[] vals = new double[IN.size()];
		int curDim;
		int numFrontDims = frontDims.size();
		double ce;
		double maxCE = -Double.MAX_VALUE;
		int maxPos = -1;
		for (int i = 0; i < numFrontDims; i++)
		{
			curDim = frontDims.get(i).intValue();
			for (int j = 0; j < vals.length; j++)
				vals[j] = dataMatrix.data.get(IN.get(j).intValue()).measures[curDim];
			
			ce = Math.abs(CRES[curDim] - computeCRE(vals, false, Constants.USE_RAO_CRE));
			if (ce > maxCE)
			{
				maxCE = ce;
				maxPos = i;
			}
		}
		
		dims.add(new Integer(frontDims.get(maxPos).intValue()));
		frontDims.remove(maxPos);
		
		return Math.max(maxCE, 0);
	}
	
	public static double computeContrastSpecial(DataMatrix dataMatrix, ArrayList<Integer> dims, ArrayList<Integer> frontDims, double[] CRES, int NumSeeds, Random ranSeedGenerator, double[] means, double[] devs) throws Exception
	{
		int num_dims = dims.size();
		int rows = dataMatrix.rows;
		
		ArrayList<Integer> dimsForClustering = new ArrayList<Integer>();
		ArrayList<Integer>[] clusters = null;
		ArrayList<Integer> curCluster = null;
		int curNumPoints;
		double[] vals = null;
		for (int i = 0; i < num_dims; i++)
			dimsForClustering.add(new Integer(dims.get(i).intValue()));
		
		clusters = incrementalClusteringNewSeed(dataMatrix, dimsForClustering, NumSeeds, ranSeedGenerator, means, devs);
		
		int curDim;
		int numFrontDims = frontDims.size();
		double ce;
		double maxCE = -Double.MAX_VALUE;
		int maxPos = -1;
		for (int i = 0; i < numFrontDims; i++)
		{
			curDim = frontDims.get(i).intValue();
			ce = 0;
			for (int j = 0; j < NumSeeds; j++)
			{
				curCluster = clusters[j];
				curNumPoints = curCluster.size();
				vals = new double[curNumPoints];
				for (int k = 0; k < curNumPoints; k++)
					vals[k] = dataMatrix.data.get(curCluster.get(k).intValue()).measures[curDim];
			
				ce += curNumPoints * computeCRE(vals, false, Constants.USE_RAO_CRE) / rows;
			} // end for
			
			ce = Math.abs(CRES[curDim] - ce);
			if (ce > maxCE)
			{
				maxCE = ce;
				maxPos = i;
			}
		}
		
		dims.add(new Integer(frontDims.get(maxPos).intValue()));
		frontDims.remove(maxPos);
		
		return Math.max(maxCE, 0);
	}
	
	public static double computeODMI(DataMatrix dataMatrix, ArrayList<Integer> dims, double CRES, int NumSeeds, Random ranSeedGenerator, double[] means, double[] devs) throws Exception
	{
		double odmi = CRES;
		int num_dims = dims.size();
		int rows = dataMatrix.rows;
		
		ArrayList<Integer> dimsForClustering = new ArrayList<Integer>();
		ArrayList<Integer>[] clusters = null;
		ArrayList<Integer> curCluster = null;
		int curNumPoints;
		double[] vals = null;
		double sum = 0;
		double partialSum;
		for (int i = 0; i < num_dims - 1; i++)
		{
			dimsForClustering.add(new Integer(i));
			clusters = incrementalClusteringNewSeed(dataMatrix, dimsForClustering, NumSeeds, ranSeedGenerator, means, devs);
			partialSum = 0;
			for (int j = 0; j < NumSeeds; j++)
			{
				curCluster = clusters[j];
				curNumPoints = curCluster.size();
				vals = new double[curNumPoints];
				for (int k = 0; k < curNumPoints; k++)
					vals[k] = dataMatrix.data.get(curCluster.get(k).intValue()).measures[dims.get(i + 1).intValue()];
			
				partialSum += curNumPoints * computeCRE(vals, false, Constants.USE_RAO_CRE) / rows;
			} // end for
			
			sum += partialSum;
		} // end for
		
		odmi = odmi - sum;
		return Math.max(odmi, 0);
	}
	
	/* useRaoCRE is true when Rao's computation of empirical CRE is used */
	public static double computeCRE(double[] vals, boolean hasSorted, boolean useGaoCRE)
	{
		int num_items = vals.length;
		if (hasSorted == false)
			Arrays.sort(vals);
		
		double cre = 0;
		double logBase = Math.log(Constants.LOG_BASE);
		if (useGaoCRE == false)
		{
			for (int i = 0; i < num_items - 1; i++)
				if (vals[i + 1] != vals[i])
					cre += (vals[i + 1] - vals[i]) * ((i + 1) / (1.0 * num_items)) * Math.log((i + 1) / (1.0 * num_items)) / logBase;
		}
		else
		{
			int count;
			for (int i = 0; i < num_items; i++)
			{
				count = 0;
				for (int j = 0; j < num_items; j++)
					if (vals[j] > vals[i])
						count++;
				
				if (count != 0)
					cre += (count / (1.0 * num_items)) * Math.log(count / (1.0 * num_items)) / logBase;
			}
		}
		
		return -cre;
	}
	
	public static ArrayList<Integer>[] incrementalClustering(DataMatrix dataMatrix, ArrayList<Integer> clusteringDims, int NumSeeds, Random ranSeedGenerator)
	{
		@SuppressWarnings("unchecked")
		ArrayList<Integer>[] ret = new ArrayList[NumSeeds];
		for (int i = 0; i < NumSeeds; i++)
			ret[i] = new ArrayList<Integer>();
		
		int rows = dataMatrix.rows;
		int width = rows / NumSeeds;
		int lower;
		int upper;
		int[] seedIds = new int[NumSeeds];
		int seedId;
		DataPoint curPoint = null;
		for (int i = 0; i < NumSeeds; i++)
		{
			lower = i * width;
			upper = (i + 1) * width - 1;
			seedId = DataProcess.generateRandomNumber(ranSeedGenerator, lower, upper, false);
			seedIds[i] = seedId;
		}
		
		DataPoint curSeed = null;
		double min;
		int minSeedId;
		double dist;
		for (int i = 0; i < rows; i++)
		{
			curPoint = dataMatrix.data.get(i);
			min = Double.MAX_VALUE;
			minSeedId = -1;
			for (int j = 0; j < NumSeeds; j++)
			{
				curSeed = dataMatrix.data.get(seedIds[j]);
				dist = DataPoint.distanceLNorm(Constants.LP_NORM, clusteringDims, curPoint, curSeed);
				if (dist < min)
				{
					min = dist;
					minSeedId = j;
				} // end if
			} // end for
			
			ret[minSeedId].add(new Integer(i));
		} // end for
		
		return ret;
	}
	
	public static ArrayList<Integer>[] incrementalClusteringNewSeed(DataMatrix dataMatrix, ArrayList<Integer> clusteringDims, int NumSeeds, Random ranSeedGenerator, double[] means, double[] devs)
	{
		@SuppressWarnings("unchecked")
		ArrayList<Integer>[] ret = new ArrayList[NumSeeds];
		for (int i = 0; i < NumSeeds; i++)
			ret[i] = new ArrayList<Integer>();
		
		int rows = dataMatrix.rows;
		int numDimsClustering = clusteringDims.size();
		DataPoint[] seeds = new DataPoint[NumSeeds];
		DataPoint curPoint = null;
		double r;
		int curDim;
		for (int i = 0; i < NumSeeds; i++)
		{
			seeds[i] = new DataPoint(numDimsClustering, 0, 0);
			r = ranSeedGenerator.nextDouble();
			for (int j = 0; j < numDimsClustering; j++)
			{
				curDim = clusteringDims.get(j).intValue();
				if (r >= 0.5)
					seeds[i].measures[j] = means[curDim] + i * devs[curDim] / (numDimsClustering * (NumSeeds - 1));
				else
					seeds[i].measures[j] = means[curDim] - i * devs[curDim] / (numDimsClustering * (NumSeeds - 1));
			}
		}
		
		DataPoint curSeed = null;
		double min;
		int minSeedId;
		double dist;
		int numMSteps = (int)Math.sqrt(rows);
		for (int i = 0; i < rows; i++)
		{
			curPoint = dataMatrix.data.get(i);
			min = Double.MAX_VALUE;
			minSeedId = -1;
			for (int j = 0; j < NumSeeds; j++)
			{
				curSeed = seeds[j];
				dist = DataPoint.distanceLNorm(Constants.LP_NORM, clusteringDims, curPoint, curSeed);
				if (dist < min)
				{
					min = dist;
					minSeedId = j;
				} // end if
			} // end for
			
			ret[minSeedId].add(new Integer(i));
			
			if (Constants.UPDATE_SEED == true && i % numMSteps == 0)
				performMStep(dataMatrix, clusteringDims, seeds, ret);
		} // end for
		
		return ret;
	}
	
	public static void performMStep(DataMatrix dataMatrix, ArrayList<Integer> clusteringDims, DataPoint[] seeds, ArrayList<Integer>[] members)
	{
		int numSeeds = seeds.length;
		int numPoints;
		DataPoint curPoint  = null;
		int numDimsClustering = clusteringDims.size();
		int curDim;
		for (int i = 0; i < numSeeds; i++)
		{
			numPoints = members[i].size();
			for (int j = 0; j < numPoints; j++)
			{
				curPoint = dataMatrix.data.get(members[i].get(j).intValue());
				for (int k = 0; k < numDimsClustering; k++)
				{
					curDim = clusteringDims.get(k).intValue();
					if (j == 0)
						seeds[i].measures[k] = curPoint.measures[curDim] / numPoints;
					else
						seeds[i].measures[k] += curPoint.measures[curDim] / numPoints;
				}
			} // end for
		} // end for
	}
}
