package duso.business;

import java.util.*;

import duso.data.*;

public class DataProcess 
{
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public static void calculateIndices(DataMatrix dataMatrix, IndexMatrix indexMatrix)
	{
		int num_rows = dataMatrix.rows;
		int num_cols = dataMatrix.cols;
		SortedObjectComparator c = new SortedObjectComparator();
		
		SortedObject[] tmp = new SortedObject[num_rows];
		for (int i = 0; i < num_rows; i++)
			tmp[i] = new SortedObject(0, 0);
		
		Constants.sortedData = new ArrayList();
		double[] vals = new double[num_rows];
		for (int j = 0; j < num_cols; j++)
		{
			for (int i = 0; i < num_rows; i++)
			{
				tmp[i].index = i;
				tmp[i].value = dataMatrix.data.get(i).measures[j];
			}
			
			Arrays.sort(tmp, c);
			for (int i = 0; i < num_rows; i++)
			{
				vals[i] = tmp[i].value;
				indexMatrix.data[i][j] = tmp[i].index;
				tmp[i].index = 0;
				tmp[i].value = 0;
			}
			Constants.sortedData.add(cumulativeDistribution(vals, true));
		}
	}
	
	public static int[] getIndexBlock(IndexMatrix indexMatrix, int colID, int startPosition, int size)
	{
		int[] ret = new int[size];
		int endPosition = startPosition + size - 1;
		for (int i = startPosition; i <= endPosition; i++)
			ret[i - startPosition] = indexMatrix.data[i][colID];
			
		return ret;
	}
	
	public static int[] getIndexBlock(int[] indexBlock, int startPosition, int size)
	{
		int[] ret = new int[size];
		int endPosition = startPosition + size - 1;
		for (int i = startPosition; i <= endPosition; i++)
			ret[i - startPosition] = indexBlock[i];
			
		return ret;
	}
	
	public static double normMinMax(double min, double max, double val)
	{
		return (val - min) / (max - min);
	}
	
	public static double normMean(double min, double max, double val)
	{
		return 2 * val / (min + max);
	}
	
	public static int generateRandomNumber(Random generator, int start, int end, boolean useFraction)
	{
		int retVal;
		if (useFraction == false)
			retVal = start + generator.nextInt(end - start + 1);
		else
		{
			long range = (long)end - (long)start + 1;
		    long fraction = (long)(range * generator.nextDouble());
		    retVal =  (int)(fraction + start);    
		}
		
		return retVal;
	}
	
	public static double[][] cumulativeDistribution(double[] a, boolean hasSorted)
	{
		if (hasSorted == false)
			Arrays.sort(a);
		int num_items = a.length;
		ArrayList<Double> distinctVals = new ArrayList<Double>();
		ArrayList<Integer> cumulativeCounts = new ArrayList<Integer>();
		for (int i = 0; i < num_items; i++)
		{
			distinctVals.add(new Double(a[i]));
			for (int j = i; j < num_items; j++)
				if (a[j] > a[i])
				{
					i = j;
					break;
				}
			
			cumulativeCounts.add(new Integer(i));
		}		
		
		int size = cumulativeCounts.size();
		double[][] ret = new double[2][size];
		for (int i = 0; i < size; i++)
		{
			ret[0][i] = distinctVals.get(i).doubleValue();
			ret[1][i] = cumulativeCounts.get(i).intValue() / (1.0 * a.length);
		}
		
		return ret;
	}
	
	public static double[] computeStatisticalMoments(double[] vals)
	{
		double sum = 0;
		for (int i = 0; i < vals.length; i++)
			sum += vals[i];
		
		sum = sum / vals.length;
		
		double variance = 0;
		for (int i = 0; i < vals.length; i++)
			variance += (vals[i] - sum)	* (vals[i] - sum);
		
		variance = Math.sqrt(variance / (vals.length - 1));
		
		double[] ret = new double[2];
		ret[0] = sum;
		ret[1] = variance;
		
		return ret;
	}
}
