package duso.data;


import duso.business.*;
import java.util.*;

public class DataPoint 
{
	public int universalID;									
	public int localID;										
	public double[] measures;								
	public double[] orgMeasures; 							
	public String[] catContexts;							
	public double[] numContexts;							
	
	public ArrayList<Integer> neighbors;					
	public ArrayList<Double> distToNeighbors;				
	public double lrd;										
	public int classID;										
	
	public DataPoint(int numMeasures, int numCatContexts, int numNumContexts)
	{
		if (numMeasures > 0)
			measures = new double[numMeasures];
		
		if (numCatContexts > 0)
			catContexts = new String[numCatContexts];
		
		if (numNumContexts > 0)
			numContexts = new double[numNumContexts];
		
		neighbors = new ArrayList<Integer>();
		distToNeighbors = new ArrayList<Double>();
	}
	
	public DataPoint(DataPoint p, int numMeasures, int numCatContexts, int numNumContexts)
	{
		this.universalID = p.universalID;
		this.localID = p.localID;
		
		if (numMeasures > 0)
			measures = new double[numMeasures];
		
		if (numCatContexts > 0)
			catContexts = new String[numCatContexts];
		
		if (numNumContexts > 0)
			numContexts = new double[numNumContexts];
		
		neighbors = new ArrayList<Integer>();
		distToNeighbors = new ArrayList<Double>();
		
		for (int i = 0; i < numMeasures; i++)
			measures[i] = p.measures[i];
		
		for (int i = 0; i < numCatContexts; i++)
			catContexts[i] = p.catContexts[i];
		
		for (int i = 0; i < numNumContexts; i++)
			numContexts[i] = p.numContexts[i];
	}
	
	public void insertNeighbors(Subspace S, DataPoint q)
	{
		double dist = DataPoint.distanceLNorm(Constants.LP_NORM, S, this, q);
		int numNeighbors = this.neighbors.size();
		int pos = numNeighbors;
		
		double tmpDist;
		for (int i = numNeighbors - 1; i >= 0; i--)
		{
			tmpDist = this.distToNeighbors.get(i).doubleValue();
			if (dist >= tmpDist)
				break;
			
			pos = i;
		}
		
		this.neighbors.add(pos, q.universalID);
		this.distToNeighbors.add(pos, dist);
		numNeighbors++;
		
		if (numNeighbors > Constants.NUM_NEIGHBORS)
		{
			if (Constants.ALLOW_TIE_NEIGHBORS == false)
			{
				for (int i = numNeighbors - 1; i >= Constants.NUM_NEIGHBORS; i--)
				{
					this.neighbors.remove(i);
					this.distToNeighbors.remove(i);
				}
			}
			else
			{
				if (this.distToNeighbors.get(Constants.NUM_NEIGHBORS - 1) != this.distToNeighbors.get(Constants.NUM_NEIGHBORS))
				{
					for (int i = numNeighbors - 1; i >= Constants.NUM_NEIGHBORS; i--)
					{
						this.neighbors.remove(i);
						this.distToNeighbors.remove(i);
					}
				}
			} // end if
		} // end if
	}
	
	public static double distanceLNorm(int expo, Subspace S, DataPoint p, DataPoint q)
	{
		int numMeasuresCols = S.num_dims;
		double dist = 0;
		int curCol;
		for (int i = 0; i < numMeasuresCols; i++)
		{
			curCol = S.dims.get(i).intValue();
			dist += Math.pow(Math.abs(p.measures[curCol] - q.measures[curCol]), expo);
		}
		
		return Math.pow(dist, 1.0 / expo);
	}
	
	public static double distanceLNorm(int expo, ArrayList<Integer> dims, DataPoint p, DataPoint q)
	{
		int numMeasuresCols = dims.size();
		double dist = 0;
		int curCol;
		for (int i = 0; i < numMeasuresCols; i++)
		{
			curCol = dims.get(i).intValue();
			dist += Math.pow(Math.abs(p.measures[curCol] - q.measures[i]), expo);
		}
		
		return Math.pow(dist, 1.0 / expo);
	}
}
